function [top_models marg_prob] = EBIR(DATA, num_results, k_i, k_e, p_i, v_0, s_0)
%************************************************************************/
%* EBIR (Exact Bayesian Inference in Regression) - A program to         */
%*    calculate the exact posterior distribution for Bayesian           */
%*    Variable Selection                                                */
%*                                                                      */
%* Please acknowledge the program authors on any publication of         */
%* scientific results based in part on use of the program and           */
%* cite the following article in which the program was described.       */
%*                                                                      */
%* E. Ruggieri and C. E. Lawrence. "On Efficient Calculations for       */
%* Bayesian Variable Selection," Computational Statistics and           */
%* Data Analysis                                                        */
%* Program Author: Eric Ruggieri                                        */
%*                                                                      */
%* Copyright (C) 2011   Duquesne University                             */
%* Duquesne University                                                  */
%* Pittsburgh, PA 15282                                                 */
%* Email:  ruggierie@duq.edu                                            */
%*                                                                      */
%* EBIR is free software: you can redistribute it and/or modify         */
%* it under the terms of the GNU General Public License as published by */
%* the Free Software Foundation, either version 3 of the License, or    */
%* (at your option) any later version.                                  */
%*                                                                      */
%* EBIR is distributed in the hope that it will be useful,              */
%* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
%* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
%* GNU General Public License for more details.                         */
%*                                                                      */
%* You should have received a copy of the GNU General Public License    */
%* along with EBIR.  If not, see <http://www.gnu.org/licenses/>         */
%* or write to the Free Software Foundation, Inc., 51 Franklin Street   */
%* Fifth Floor, Boston, MA  02110-1301, USA.                            */
%************************************************************************/
%
%
% Outline of the code is as follows:
% 1) Define default parameter values if not specified by user
% 2) Initialize Variables to be used throughout the program
% 3) EBIR Recursion - actual function at end of script
% 4) Processing the Results
%
% Description of Parameters, input, and output can be found in accompanying 
% ReadMe file
%

%For function syntax, enter EBIR with no arguments.  
%For examples, enter EBIR('examples');

%*******Examples*******
%
if(nargin ==1)
    if(isstr(DATA))
        if (strcmp(DATA,'examples'))
        disp('  ')
        disp(' Example 1: Random Data - Return the entire posterior space')
        disp('  X=randn(100,10); %10 Variables, 100 observations')
        disp('  Y = 4*X(:,3) - 3*X(:,8) + randn(100,1); ')
        disp('  [top_models marg_prob] = EBIR([Y X]);')
        disp('  ')
        disp(' Example 2: Random Data - Return only the top 10 sub-models')
        disp('  X=randn(100,10); %10 Variables, 100 observations')
        disp('  Y = 4*X(:,3) - 3*X(:,8) + randn(100,1); ')
        disp('  [top_models marg_prob] = EBIR([Y X], 10);')
        disp('  ')
        disp(' Example 3: The Crime and Punishment Data Set - Default Parameter Values')
        disp('  load(''CRIME_DATA.txt'');')
        disp('  Y=log(CRIME_DATA(:,1));   %log transform and standardize the data')
        disp('  Y=(Y-mean(Y))/std(Y);')
        disp('  X=CRIME_DATA(:,2:16);')
        disp('  X(:,1)=log(X(:,1));')
        disp('  X(:,1)=(X(:,1)-mean(X(:,1)))/std(X(:,1));')
        disp('  X(:,2)=(X(:,2)-mean(X(:,2)))/std(X(:,2));  %Indicator Variable, so not log transformed')  
        disp('  for i=3:15      %15 Variables in the Crime and Punishment Data Set')
        disp('     X(:,i)=log(X(:,i));')
        disp('     X(:,i)=(X(:,i)-mean(X(:,i)))/std(X(:,i));')
        disp('  end')
        disp('  [top_models marg_prob] = EBIR([Y X]);')
        disp('  ')
        disp(' Example 4: The Crime and Punishment Data Set - Changing the Default Parameter Values')
        disp('  load(''CRIME_DATA.txt'');')
        disp('  Y=log(CRIME_DATA(:,1));   %log transform and standardize the data')
        disp('  Y=(Y-mean(Y))/std(Y);')
        disp('  X=CRIME_DATA(:,2:16);')
        disp('  X(:,1)=log(X(:,1));')
        disp('  X(:,1)=(X(:,1)-mean(X(:,1)))/std(X(:,1));')
        disp('  X(:,2)=(X(:,2)-mean(X(:,2)))/std(X(:,2));  %Indicator Variable, so not log transformed')  
        disp('  for i=3:15      %15 Variables in the Crime and Punishment Data Set')
        disp('     X(:,i)=log(X(:,i));')
        disp('     X(:,i)=(X(:,i)-mean(X(:,i)))/std(X(:,i));')
        disp('  end')
        disp('  [top_models marg_prob] = EBIR([Y X], 25, 0.1, 10, 0.75, 2, 1);')
        return
        else
            disp('[top_models marg_prob] = EBIR(DATA, num_results, k_i, k_e, p_i, v_0, s_0)')
            disp('For examples, enter EBIR(''examples'')')
            return
        end
    end
end
        
%*******Check for Inputs*********
if(nargin==0)
    disp('[top_models marg_prob] = EBIR(DATA, num_results, k_i, k_e, p_i, v_0, s_0)')
    disp('For examples, enter EBIR(''examples'')')
    return
end
%******************Load the Data**************************
[N num_vars] = size(DATA);  %N is the total number of observations
Y=DATA(:,1); X=DATA(:, 2:num_vars);
num_vars=num_vars-1;        %Since the first column contains Y
%*********Set Default Parameter Values********************
if (nargin<2)
    num_results=0;          %Default is to return the entire posterior space
end

if (nargin <4)
    k_i=0.01;               %INCLUDE
    k_e=100;                %EXCLUDE 
end
%We first exclude all models - i.e. X'*X +k_e*I and then add in the variables
%one at a time.  When we update, we add the quantity (k_i-k_e) 
k_chg=k_i-k_e;              %The amount by which we have to update the matrix.

if(nargin<5)
    p_i = 0.5;              %p_i is the prior probability of including a variable in the regression
end
p_e=1-p_i;                  %p_e is the prior probability of excluding a variable from the regression
p_e=p_e^2; p_i=p_i^2;       %As opposed to squaring these terms for each probability calculation

if(nargin<7)
    v_0=5; s_0=0.1;         %prior parameters for inverse chi-square distribution
end
v_n=v_0+N;                  %posterior degrees of freedom

%*****************Create Initial Matrices***************************

XTX=X'*X; yTy=Y'*Y; XTy=X'*Y; 
G=XTX+k_e*eye(num_vars); 
Ck=inv(G); detG=log(det(G));
Am=zeros(1,num_vars);       %Vector indicating which sub-model

if(num_results==0)          %Look at the full posterior space
    top_models=zeros(1,2^num_vars);
else                        %Look at only the top 'num_results'
    top_models=zeros(num_results, num_vars+1);
    top_models(:,1)=-Inf;
end
    
marg_prob=zeros(1,num_vars)-Inf;    %Marginal Probability of Each Variable being Included
total_prob=-Inf;                    %Normalization Constant
count=1;                            %used as index for the full posterior space

%*****************Recursion************************************

[top_models marg_prob total_prob]=tree(top_models, marg_prob, total_prob, count, Ck, detG, Am, 0);

%*****************Processing Results***************************

if(num_results>0)           %Further processing of top sub-models
    top_models(:,1)= -1*exp(top_models(:,1)-total_prob);    %Normalize
    top_models=sortrows(top_models); top_models(:,1)=-1*top_models(:,1);
                            %To sort descending, multiply by -1 twice
    for i=1:num_results
        top_models(i,2:num_vars+1)=top_models(i,2:num_vars+1).*(1:num_vars);    
                            %Replace Indicator Variables with Variable Numbers
    end
else
    top_models(:)=exp(top_models(:)-total_prob);    %Normalize 
end

marg_prob(:)=exp(marg_prob(:)-total_prob);          %Normalize marginal probabilities

%*****************Recursive Function tree**********************
function [top_models marg_prob total_prob count] = tree(top_models, marg_prob,total_prob, count,Ck, detG, Am, k)
        k=k+1;
        if (k<num_vars)
            %Exclusion = no update
            [top_models marg_prob total_prob count] = tree(top_models, marg_prob,total_prob,count,Ck, detG, Am,k);
            %Inclusion = update our matrix
            Am(k)=1;                    %include variable k
            E=Ck(:,k)*k_chg; tr=E(k);   %E = Ck*Ek; tr = trace(Ck*Ek)
            Ck=Ck-E*Ck(k,:)/(1+tr);     %Update Inverse
            detG=log(1+tr)+detG;        %Update Determinant
            
            [top_models marg_prob total_prob count]=tree(top_models, marg_prob,total_prob,count,Ck, detG, Am,k);
            
        else
            %The Final Step in the Recursion - Calculate the Probability
            %Exclusion = no update
            m_i=sum(Am);                %m_i= # included variables
            hat_beta=Ck*XTy;            %regression coefficients
            I_Am= Am; 
            I_Am=I_Am*k_chg+k_e;        %Diagonal Matrix
            
            %lambda =(Y-X*hat_beta)'*(Y-X*hat_beta)+I_Am.*hat_beta'*hat_beta+v_0*s_0;  
            lambda=yTy - 2*hat_beta'*XTy + hat_beta'*XTX*hat_beta+I_Am.*hat_beta'*hat_beta+ v_0*s_0;
            % Equivalent Statements, speed of each depends on the data set
            P1 = (m_i/2)*log(k_i*p_i)+ (num_vars-m_i)/2*log(k_e*p_e) -v_n/2*log(lambda)-0.5*detG;
            
            %Inclusion = update our matrix
            Am(k)=1;                    %Include the final variable
            E=Ck(:,k)*k_chg; tr=E(k);   %E = Ck*Ek; tr = trace(Ck*Ek)
            Ck=Ck-E*Ck(k,:)/(1+tr);     %update the inverse
            detG=log(1+tr)+detG;        %calculate Determinant
            
            m_i=sum(Am);                %Total Number of included variables
            hat_beta=Ck*XTy;            %Regression Coefficients
            I_Am = Am;
            I_Am = I_Am*k_chg+k_e;      %Diagonal Matrix
            
            %lambda = (Y-X*hat_beta)'*(Y-X*hat_beta) + I_Am.*hat_beta'*hat_beta + v_0*s_0;
            lambda=yTy - 2*hat_beta'*XTy + hat_beta'*XTX*hat_beta+I_Am.*hat_beta'*hat_beta+ v_0*s_0;
            % Equivalent Statements, speed of each depends on the data set
            P2 = (m_i/2)*log(k_i*p_i)+ (num_vars-m_i)/2*log(k_e*p_e) -v_n/2*log(lambda)-0.5*detG;
            
            %Store the Results:
            if(num_results>0)           %Check if current sub-model is one of the 'num_results' best
                [M index] = min(top_models(:,1));
                if(P1> M)
                    top_models(index,1)=P1;             %Record Probability
                    top_models(index,2:num_vars+1)=Am;  %Record Sub-Model
                    top_models(index, num_vars+1)=0;    %Am(num_vars) was change to calculate P2
                end
                [M index] = min(top_models(:,1));
                if(P2> M)
                    top_models(index,1)=P2;
                    top_models(index,2:num_vars+1)=Am;
                end
            else                        %Place among others in full description of posterior space
                top_models(count)=P1; top_models(count+1)=P2;
            end
            
            count=count+2;              %Two Sub-Models Calculated
            
            w=addtwo([P2 P1]);
            %Calculate Normalization Constant
            if(total_prob>-Inf)         
                total_prob=addtwo([total_prob w]);
            else
                total_prob=w;
            end
            %Calculate Marginal Probabilities
            for i=1:num_vars-1      
               if(Am(i)==1)             %If Variable Included
                   if(marg_prob(i)>-Inf)
                       marg_prob(i) = addtwo([marg_prob(i) w]);
                   else
                       marg_prob(i) =w;
                   end
               end
            end
            if(marg_prob(num_vars)>-Inf)
                marg_prob(num_vars) = addtwo([marg_prob(num_vars) P2]);
            else
                marg_prob(num_vars) =P2;
            end
            
        end
        
end
%*****************Function AddTwo********************
%Adds two numbers in log form and stores the result as a logarithm
function w = addtwo(v)
w = v(1)+log(1+exp(v(2)-v(1))); 
end


%****************End of EBIR****************
end